import hashlib
import hmac
import json
import time
import os
import requests
from pathlib import Path
from datetime import datetime
from io import BytesIO
import configparser
from requests.exceptions import ChunkedEncodingError

from PIL import Image

from mayak_api.data_geo_object import DataGeoObject
from mayak_api.data_review import DataReview
from mayak_api.config_logger import BaseLogger


base_logger = BaseLogger()
base_logger.path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'mayak_api_client.log')
logger = base_logger.create_stream()


class MayakApiClient:
    status: str = 'test'  # 'prod'/'test'
    mayak_api_secret: str = None
    partner_id: int = 1
    id_for_headers: str = '20'
    priority_for_image: int = 50

    def __init__(self):
        config_dict = self.config('mayak_api', 'config.ini')
        self.mayak_api_secret = config_dict['mayak_api_secret']

    @staticmethod
    def config(file_dir: str, file_name: str) -> dict:
        """COMMON Метод осуществляет парсинг конфигурационного файла и возвращает словарь
           :param str file_dir: название директории файлов
           :param str file_name: название файла
        """
        common_dir = Path(os.path.dirname(os.path.realpath(__file__))).parent
        need_path = os.path.join(common_dir, file_dir, file_name)
        result = dict()
        config = configparser.ConfigParser()
        config.read(need_path)
        auth, data, mode = config['AUTH'], config['DATA'], config['MODE']
        for key, value in auth.items():
            result.update({key: value})
        for key, value in data.items():
            result.update({key: value})
        for key, value in mode.items():
            result.update({key: value})
        return result

    @staticmethod
    def make_request_with_retry(url: str, retries: int = 5, **kwargs) -> requests.Response | bool:
        """
        Осущевстляет получение данных страницы, расположенной по url, в случае ошибки
        повторяет подключение через 5 секунд в количестве 5 раз
        :param url: домен для подключения
        :param retries: кол-во повторных попыток (по умолчанию = 5)
        :param kwargs: иные параметры необходимые для requests.get
        :return: результат запроса или False
        """

        for _ in range(retries):
            try:
                response = requests.get(url, **kwargs)
                return response
            except ChunkedEncodingError as e:
                logger.error(f"Ошибка ChunkedEncodingError: {e}")
                logger.info("Повторная отправка запроса через 5 секунд...")
                time.sleep(5)

        logger.warning("Не удалось выполнить запрос после нескольких попыток")
        return False

    def _get_hash(self, hash_string: str) -> hmac:
        """
        Метод хеширования для маяка.
        :param hash_string: в строке должны быть следующие величины [url, метод (POST, GET), str(round(unixtime)), data_string]
        :return: хеш
        """
        return hmac.new(self.mayak_api_secret.encode(), hash_string.encode(), hashlib.sha256).hexdigest()

    @staticmethod
    def params_handler(params: dict) -> str:
        """
        Обрабатывает параметры запроса для хеширования
        """
        val_list = []
        for key, value in params.items():
            val_list.append(f'{key}={value}')
        val_list.sort()
        return ';'.join(val_list)

    def make_request(self, url: str, type_connection: str,
                     **kwargs) -> bool | requests.Response:
        """
        Создает подключение к api по методу POST

        :param dict_data: данные для отправки
        :param url: url API
        :param type_connection
        :param type_api: object/gallery/media
        :param kwargs:
        :return: в случае ошибки False
        """
        unixtime = time.mktime(datetime.now().timetuple())
        if type_connection == 'POST':

            data_value = json.dumps(kwargs.get('dict_data'), default=str)
            data_string = f'{kwargs.get("type_api")}=' + data_value

            data = {f'{kwargs.get("type_api")}': data_value}
            del kwargs['dict_data']
            del kwargs['type_api']
            hash_string = ';'.join([url, type_connection, str(round(unixtime)), data_string])
            headers = {
                'id': self.id_for_headers,
                'hash': self._get_hash(hash_string),
                'timestamp': str(round(unixtime)),
            }
            resp = requests.post(url=url, headers=headers, data=data, **kwargs)
        elif type_connection == 'GET':

            if kwargs.get('params_for_handler'):
                hash_string = ';'.join(
                    [url, type_connection, str(round(unixtime)), kwargs['params_for_handler']]
                )
            else:
                hash_string = ';'.join([url, type_connection, str(round(unixtime))])

            headers = {
                'id': self.id_for_headers,
                'hash': self._get_hash(hash_string),
                'timestamp': str(round(unixtime)),
            }
            resp = requests.get(url=url, headers=headers)
        else:
            hash_string = ';'.join([url, type_connection, str(round(unixtime))])
            headers = {
                'id': self.id_for_headers,
                'hash': self._get_hash(hash_string),
                'timestamp': str(round(unixtime)),
            }
            resp = requests.delete(url=url, headers=headers)

        if resp.status_code == 200:
            try:
                if resp.json()['status'] == 'success':
                    return resp.json()
            except Exception as error:
                logger.error(f'{error=}')
                return False
        elif resp.status_code == 500:
            logger.error(f'Запрос к API c {url=} вернул {resp.status_code=} cо следующим содержимым {resp.json()}')
            return False
        else:
            logger.error(f'Запрос к API c {url=} вернул {resp.status_code=}')
            return False

    # region Работа с geo_object
    def object_load_object_post(self, data: DataGeoObject) -> bool:
        """
        :param data: передается объет класса Data, который может содержать следующие свойства
            object_type_id: int
            title_obj: str
            description: str
            description_full_obj: str
            lat: float
            lon: float
            address: str
            city_id: int
            phone: str
            email: str
            site: str
            instagram: str
            vk: str
            telegram: str
            facebook: str

        :return: возвращает False в случае ошибки
        """
        get_dict_data: dict = data.create_correct_dict()
        url = f'http://object.mayak.{self.status}/api/v1/object'

        if not get_dict_data.get('address'):
            return False

        if not get_dict_data.get('city_id'):
            return False

        if len(get_dict_data['address']) > 127:
            get_dict_data['address'] = get_dict_data['address'].split(',')[0]

        data_dict_for_load = {
            "object_type": get_dict_data['object_type_id'],
            "title": get_dict_data['title_obj'],
            "description": get_dict_data['description'],
            "description_full": get_dict_data['description_full_obj'],
            "city_id": get_dict_data['city_id'],
            "lat": get_dict_data['lat'],
            "lon": get_dict_data['lon'],
            "address": get_dict_data['address'],
            "worktime": get_dict_data['worktime'],
            "phone": get_dict_data['phone'],
            "email": get_dict_data['email'],
            "site": get_dict_data['site'],
            "instagram": get_dict_data['instagram'],
            "vk": get_dict_data['vk'],
            'telegram': get_dict_data['telegram'],
            'facebook': get_dict_data['facebook'],
            "language": get_dict_data['language'],
            "status": get_dict_data['status_obj'],
            "partner_id": self.partner_id,
        }

        flag_without_error = self.make_request(url, 'POST', type_api='object', dict_data=data_dict_for_load)

        if not flag_without_error:
            logger.error(f'Загрузка объекта в Маяк не прошла')
            return False
        else:
            data.object_id = flag_without_error.json()['result']['id']

    def object_update_object_post(self, data: DataGeoObject) -> bool:
        """
        ПОзволяет обновить объекта с заданным id
        ::param data: передается объет класса Data, который должен обязательно содержать следующие свойства
            object_id: int

            а также свойства которые необходимо обновить у указанного объекта \n
            (все свойства = None, буду проигнорированы)
                        object_type_id: int \n
                        title_obj: str \n
                        description: str \n
                        description_full_obj: str\n
                        lat: float\n
                        lon: float\n
                        address: str\n
                        city_id: int\n
                        phone: str\n
                        email: str\n
                        site: str\n
                        instagram: str\n
                        vk: str\n
                        telegram: str\n
                        facebook: str\n

        :return:
        """

        dict_for_update = {k: v for k, v in data.create_correct_dict().items() if v is not None}

        if not dict_for_update.get('object_id'):
            logger.error('Забыли задать object_id')
            return False

        url = f'http://object.mayak.{self.status}/api/v1/object/{dict_for_update["object_id"]}'
        flag_without_error = self.make_request(url, 'POST', type_api='object', dict_data=dict_for_update)

        if not flag_without_error:
            logger.error(f'Оновление объекта в Маяк не прошлло')
            return False
        else:
            logger.info('Обновление прошло успешно')

    # endregion

    # region Работа с media_object
    def media_create_gallery_post(self, data: DataGeoObject) -> bool:
        """
        Создаёт галерею. Добавляет id галерии к свойствам объекта

        :param data: передается объет класса Data, который должен обязательно содержать следующие свойства
                title_obj: str
                object_id: int
                object_type_gallery: str = 'tourproduct'
                awards_gallery: str = ''

        :return: в случае ошибки - False
        """
        get_dict_data: data = data.create_correct_dict()

        if not get_dict_data.get('title_obj'):
            logger.error(f'Вы забыли задать свойство объкта -  title_obj')
            return False
        elif len(get_dict_data['title_obj']) > 63:
            title_gallery = get_dict_data['title_obj'][:62]
            description_full_gallery = get_dict_data['title_obj']
        else:
            title_gallery = get_dict_data['title_obj']
            description_full_gallery = None

        if not get_dict_data.get('object_id'):
            logger.error(f'Вы забыли задать свойство объкта -  object_id')
            return False

        post_data = {
            'title': title_gallery,
            'description_full': description_full_gallery,
            'awards': get_dict_data['awards_gallery'],
            'partner_id': self.partner_id,
            'object_type': get_dict_data['object_type_gallery'],
            'object_id': get_dict_data['object_id']
        }

        url = f'http://media.mayak.{self.status}/api/v1/gallery'

        flag_without_error = self.make_request(url, 'POST', type_api='gallery', dict_data=post_data)

        if not flag_without_error:
            logger.error(f'Загрузка объекта в Маяк не прошла')
            return False
        else:
            data.media_gallery_id = flag_without_error.json()['result']['id']

    def media_load_image_post(self, data: DataGeoObject, file_url: str, description: str = None) -> bool:
        """
        Выполняет загрузку фотографии по АПИ
        :param data: передается объет класса Data, который должен обязательно содержать следующие свойства
                media_gallery_id: int
                object_id: int
                object_type_gallery: str = 'tourproduct'

        :param file_url: url по которой находится картинка
        :param description: описание картинки
        :return: в случае ошибки False
        """
        get_dict_data: data = data.create_correct_dict()

        if not (get_dict_data.get('object_type_gallery')
                or get_dict_data.get('object_id')
                or get_dict_data.get('media_gallery_id')
        ):
            logger.error('Проверьте свойства, котоыре должны быть обязательны для передаваемого объекта')
            return False

        post_data = {
            'type': 1,
            'title': 'Picture',
            'partner_id': self.partner_id,
            'object_type': get_dict_data['object_type_gallery'],
            'object_id': get_dict_data['object_id'],
            'file_url': file_url,
            'gallery_id': get_dict_data['media_gallery_id'],
            'priority': self.priority_for_image,
            'description': description
        }
        url = f'http://media.mayak.{self.status}/api/v1/media'

        req = self.make_request_with_retry(
            file_url,
            headers={
                'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/86.0.4240.198 Safari/537.36'},
        )
        logger.info(f'Картинка имеет статус - {req.status_code}')
        try:
            image = Image.open(BytesIO(req.content))
        except Exception:
            logger.error(Exception)
            image = None

        # иногда возникает ошибка, для этого выполняется дополнительная проверка на формат
        if image and image.format == 'JPG':
            logger.info(image.format)
            image.save(BytesIO(req.content), 'jpeg')

        files = {'file': req.content}

        flag_without_error = self.make_request(url, 'POST', type_api='media', dict_data=post_data, files=files)

        if not flag_without_error:
            logger.error(f'Загрузка картинки в Маяк c {file_url} не прошла')
            return flag_without_error

    def media_get_gallery_get(self, gallery_id: int):
        """
        Получаем сведения о галерии из таблици media.media_gallery по его id
        :param gallery_id: id галереи
        :return:
        """
        url = f'http://media.mayak.{self.status}/api/v1/gallery/{gallery_id}'
        response: bool | dict = self.make_request(url, 'GET')
        return response['result'] if response else response

    def media_get_image_get(self, image_id: int): # узнать
        """
        тут мы поулчаем не картинку, а хз как назвать - надо узнать
        Получаем сведения о галерии из таблици media.media по его id
        :param image_id:
        :return:
        """
        url = f'http://media.mayak.{self.status}/api/v1/media/{image_id}'
        response: bool | dict = self.make_request(url, 'GET')
        return response['result'] if response else response

    # endregion

    # region Работа с review_object
    def like_rating_get_review_get(self, review_id: int) -> bool | dict:
        """
        Получение конкретного отзыва из таблицы like_rating.like_rating_review
            в случае успешности возвращает словарь содержащий всю информацию об отзыве
            (копия строки таблицы) \n
            в случае ошибки возвращает False \n

        :param review_id: id отзыва
        :return: или отзыв или False
        """
        url = f'http://like-rating.mayak.{self.status}/api/v1/review/{review_id}'
        response: bool | dict = self.make_request(url, 'GET')
        return response['result'] if response else response

    def like_rating_create_review_post(self, data: DataReview) -> bool | dict:
        """
            :param data: передается объет класса DataReview,
            который должен содержать следующие свойства
            ['type', 'uid', 'status', 'platform_id', 'external_link',
                           'reviewer_email', 'title', 'description', 'external_rating',
                           'rating', 'like_count', 'dislike_count', 'review_platform_id',
                           'review_date_time', 'review_date_time']

        :return: возвращает False в случае ошибки

        """

        url = f'http://like-rating.mayak.{self.status}/api/v1/review'
        data_with_dict = data.create_correct_dict()

        required_fields = ['type', 'uid', 'status', 'platform_id', 'external_link',
                           'reviewer_email', 'title', 'description', 'external_rating',
                           'rating', 'like_count', 'dislike_count', 'review_platform_id',
                           'review_date_time', 'review_date_time']

        missing_fields = [field for field in required_fields if not data_with_dict.get(field)]
        if missing_fields:
            raise KeyError(f"Отсутствуют обязательные для загрузки данные в DataReview: "
                           f"{', '.join(missing_fields)}"
                           )

        response: bool | requests.Response = self.make_request(url, 'POST',
                                                               type_api='review', dict_data=data_with_dict)
        if not response:
            logger.error(f'Загрузка объекта в Маяк не прошла')
            return False
        else:
            data.review_id = response.json()['result']['id']

    def like_rating_update_review_post(self, data: DataReview):
        """
        Позволяет обновить отзыв, при этом необходимо передать все следующие свойста
        ['type', 'uid', 'status', 'platform_id', 'external_link',
                           'reviewer_email', 'title', 'description', 'external_rating',
                           'rating', 'like_count', 'dislike_count', 'review_platform_id',
                           'review_date_time', 'review_date_time']
        :param data: объект класса
        :return:
        """
        data_with_dict = data.create_correct_dict()

        dict_for_update = {k: v for k, v in data.create_correct_dict().items() if v is not None}

        if not dict_for_update.get('review_id'):
            logger.error('Забыли задать review_id')
            return False

        url = f'http://like-rating.mayak.{self.status}/api/v1/review/{dict_for_update['review_id']}'

        response: bool | requests.Response = self.make_request(url, 'POST',
                                                               type_api='review', dict_data=data_with_dict)
        if not response:
            logger.error(f'Загрузка объекта в Маяк не прошла')
            return False
        else:
            logger.info('Обновление прошло успешно')

    def like_rating_delete_review_delete(self, data: DataReview):

        data_with_dict = data.create_correct_dict()
        if not data_with_dict.get('review_id'):
            logger.error('Забыли задать review_id')
            return False

        url = f'http://like-rating.mayak.{self.status}/api/v1/review/{data_with_dict['review_id']}'

        response: bool | requests.Response = self.make_request(url, 'DELETE')
        if not response:
            logger.error(f'Удаление объекта из Маяк не получилось')
            return False
        else:
            logger.info('Удаление прошло успешно')

    def like_rating_get_list_review_get(self, type_object: str,
                                        uid_object: int): # нужно протестить
        """
        Получает список отзывов для конкретного объекта
        :param type_object: свойство type класса DataReview
        :param uid_object: свойство uid класса DataReview
        :return:
        """
        url = f'http://like-rating.mayak.{self.status}/api/v1/review/list'

        param = {
            'uid': uid_object,
            'type': type_object,
        }

        params_for_handler = self.params_handler(param)

        response: bool | dict = self.make_request(url, 'GET', params_for_handler=params_for_handler)
        return response['result'] if response else response
    # endregion
